/*
 *	linux/fs/rbf/dir.c
 *	RBF File System - directory handling functions
 *
 *	Copyright (C) 2001 Andrew Cannon <ajc@gmx.net>
 *
 * from
 *  
 *  linux/fs/ext2/dir.c
 *  
 * Copyright (C) 1992, 1993, 1994, 1995
 * Remy Card (card@masi.ibp.fr)
 * Laboratoire MASI - Institut Blaise Pascal
 * Universite Pierre et Marie Curie (Paris VI)
 *
 * from
 *
 *  linux/fs/minix/dir.c
 *
 * Copyright (C) 1991, 1992  Linus Torvalds
 *
 *  This file is subject to the terms and conditions of the
 *  GNU General Public License. See the file COPYING in the
 *  main directory of this archive for more details.
 */

#include "rbf_defs.h"
#include "rbf_page.h"

#undef DEBUG
#define DEBUG

#ifndef DEBUG
#undef DPRINTK
#define DPRINTK(x...)
#endif

static inline void rbf_put_page(struct page *page)
{
	kunmap(page);
	page_cache_release(page);
}

static inline int rbf_commit_chunk(struct page *page, unsigned from, unsigned to)
{
	return (block_commit_write(page, from, to));
}

static inline void rbf_check_page(struct page *page)
{
	/* no checking yet... */
	SetPageChecked(page);
}

static struct page *rbf_get_page(struct inode *dir, unsigned long n)
{
	struct address_space *mapping = dir->i_mapping;
	struct page *page = read_cache_page(mapping, n, (filler_t*)mapping->a_ops->readpage, NULL);

	if (IS_ERR(page))
		goto sync_fail;
	wait_on_page_locked(page);
	kmap(page);
	if (!PageUptodate(page))
		goto async_fail;
	if (!PageChecked(page))
		rbf_check_page(page);
	if (PageError(page))
		goto async_fail;
	return(page);

async_fail:
	rbf_put_page(page);
	return ERR_PTR(-EIO);

sync_fail:
	return(page);
}

static int rbf_readdir(struct file *f, void *dirent, filldir_t filldir)
{
	struct inode *dir = f->f_dentry->d_inode;
	struct super_block *s = dir->i_sb;
	struct page *page;
	struct rbf_dirent *de;
	unsigned int offset;
	char *kaddr;
	char name[RBF_NAMELEN];

	DPRINTK("in pos=%d\n", (int)f->f_pos);
	
	if (f->f_pos & (RBF_DIRENT_SIZE-1)) {
		printk(KERN_ERR "Bad f_pos=%08lx for %s:%08lx\n", (unsigned long)f->f_pos, 
			s->s_id, dir->i_ino);
		return -EBADF;
	}

	lock_kernel();

	while (f->f_pos < dir->i_size) {
		offset = f->f_pos & (PAGE_CACHE_SIZE-1);
		page = rbf_get_page(dir, f->f_pos >> PAGE_CACHE_SHIFT);
		
		if (IS_ERR(page)) {
			f->f_pos += PAGE_CACHE_SIZE - offset;
			continue;
		}
		kaddr = page_address(page);
		do {
			de = (struct rbf_dirent *)(kaddr + offset);
			if (de->dir_name[0]) {
				int size = rbf_readname(name, de->dir_name);
				ino_t addr = rbf_read(de->dir_addr, 3);
				
				DPRINTK("entry '%.28s' addr %08lx\n", name, addr);
				if (filldir(dirent, name, size, f->f_pos, addr, DT_UNKNOWN) < 0) {
					rbf_put_page(page);
					unlock_kernel();
					return 0;
				}
			}
			offset += RBF_DIRENT_SIZE;
			f->f_pos += RBF_DIRENT_SIZE;
		}
		while (offset < PAGE_CACHE_SIZE && f->f_pos < dir->i_size);
		rbf_put_page(page);
	}

	DPRINTK("out\n");

	unlock_kernel();

	return 0;	
}

struct file_operations rbf_page_dir_operations = {
	.read =		generic_read_dir,
	.readdir =	rbf_readdir,
//	.ioctl =	rbf_ioctl,
//	.fsync =	rbf_sync_file,
	.fsync =	file_fsync
};



/*
 *	rbf_find_entry()
 *
 * finds an entry in the specified directory with the wanted name. It
 * returns the page in which the entry was found, and the entry itself.
 * Page is returned mapped and unlocked.
 * Entry is guaranteed to be valid.
 */
struct rbf_dirent *rbf_find_entry(struct inode *dir, struct dentry *dentry, struct page **res_page)
{
	struct super_block *s = dir->i_sb;
	const char *name = dentry->d_name.name;
	int namelen = dentry->d_name.len;
	unsigned long pos, offset;
	struct page *page;
	struct rbf_dirent *de;
	char *kaddr;

	DPRINTK("name=%s, namelen=%d\n", name, namelen);
	
	if (namelen > RBF_NAMELEN)
		return NULL;
	
	pos = 0;
	while (pos < dir->i_size) {
		offset = pos & (PAGE_CACHE_SIZE-1);
		page = rbf_get_page(dir, pos >> PAGE_CACHE_SHIFT);
		
		if (IS_ERR(page)) {
			pos += PAGE_CACHE_SIZE - offset;
			continue;
		}
		kaddr = page_address(page);
		do {
			de = (struct rbf_dirent *)(kaddr + offset);
			if (de->dir_name[0] && rbf_match(s, namelen, name, de->dir_name)) {
				*res_page = page;
				DPRINTK("out entry %ld\n", pos / RBF_DIRENT_SIZE);
				return de;
			}
			offset += RBF_DIRENT_SIZE;
			pos += RBF_DIRENT_SIZE;
		}
		while (offset < PAGE_CACHE_SIZE && pos < dir->i_size);
		rbf_put_page(page);
	}

	DPRINTK("out 0\n");
	*res_page = 0;
	return 0;
}

struct rbf_dirent *rbf_dotdot(struct inode *dir, struct page **res_page)
{
	struct dentry d;		/* dummy for name lookup */
	
	d.d_name.name = "..";
	d.d_name.len = 2;
	return rbf_find_entry(dir, &d, res_page);
}

void rbf_release_entry(struct page *page)
{
	rbf_put_page(page);
}

ino_t rbf_inode_by_name(struct inode *dir, struct dentry *dentry)
{
	ino_t ino = 0;
	struct rbf_dirent *de;
	struct page *page;
	
	de = rbf_find_entry(dir, dentry, &page);
	if (de) {
		ino = rbf_read(de->dir_addr, 3);
		rbf_put_page(page);
	}
	return ino;
}

/* Releases the page */
void rbf_set_link(struct inode *dir, struct rbf_dirent *de, struct page *page, struct inode *inode)
{
	unsigned from = (char *)de - (char *)page_address(page);
	unsigned to = from + RBF_DIRENT_SIZE;
	int err;

	lock_page(page);
	err = page->mapping->a_ops->prepare_write(NULL, page, from, to);
	if (err)
		BUG();
	rbf_write(de->dir_addr, 3, inode->i_ino);
	err = rbf_commit_chunk(page, from, to);
	unlock_page(page);
	rbf_put_page(page);
	dir->i_mtime = dir->i_ctime = CURRENT_TIME;
	mark_inode_dirty(dir);
}

/*
 *	Parent is locked.
 */
int rbf_add_link(struct dentry *dentry, struct inode *inode)
{
	struct inode *dir = dentry->d_parent->d_inode;
	struct super_block *s = dir->i_sb;
	const char *name = dentry->d_name.name;
	int namelen = dentry->d_name.len;
	struct page *page;
	struct rbf_dirent *de;
	unsigned long pos = 0;
	int err;
	char *kaddr;
	unsigned offset, from, to;

	DPRINTK("name=%s, namelen=%d\n", name, namelen);

	err = rbf_validate_name(s, name, namelen);
	if (err)
		return err;

	while (pos <= dir->i_size) {
		offset = pos & (PAGE_CACHE_SIZE-1);
		page = rbf_get_page(dir, pos >> PAGE_CACHE_SHIFT);
		
		if (IS_ERR(page)) {
			pos += PAGE_CACHE_SIZE - offset;
			continue;
		}
		kaddr = page_address(page);
		do {
			de = (struct rbf_dirent *)(kaddr + offset);
			if (de->dir_name[0] == 0)
				goto got_it;
			err = -EEXIST;
			if (rbf_match(s, namelen, name, de->dir_name))	/* is this necessary? */
				goto out_page;
			offset += RBF_DIRENT_SIZE;
			pos += RBF_DIRENT_SIZE;
		}
		while (offset < PAGE_CACHE_SIZE && pos <= dir->i_size);
		rbf_put_page(page);
	}
	/* reached end of dir - add a new entry */

	err = rbf_find_sector(dir, pos >> s_su_sectsize_bits, 0, 1);
	if (err < 0)
		goto out;
	offset = pos & (PAGE_CACHE_SIZE-1);
	page = rbf_get_page(dir, pos >> PAGE_CACHE_SHIFT);
	err = -EIO;
	if (IS_ERR(page))
		goto out;
	kaddr = page_address(page);
	de = (struct rbf_dirent *)(kaddr + offset);
	dir->i_size += RBF_DIRENT_SIZE;				/* expand directory */

got_it:
	from = (char*)de - (char*)page_address(page);
	to = from + RBF_DIRENT_SIZE;
	
	DPRINTK("free entry %ld page %08lx from %d to %d mapping %08lx\n",
		pos / RBF_DIRENT_SIZE, (unsigned long)page, from, to, (unsigned long)page->mapping);
	
	lock_page(page);
	err = page->mapping->a_ops->prepare_write(NULL, page, from, to);
	if (err)
		goto out_unlock;
		
	strncpy(de->dir_name, name, RBF_NAMELEN + 1);		/* clear dir_res1 as well */
	de->dir_name[namelen - 1] |= 0x80;
	rbf_write(de->dir_addr, 3, inode->i_ino);
		
	err = rbf_commit_chunk(page, from, to);
	dir->i_mtime = dir->i_ctime = CURRENT_TIME;
	mark_inode_dirty(dir);
	/* OFFSET_CACHE */
out_unlock:
	unlock_page(page);
out_page:
	rbf_put_page(page);
out:
	DPRINTK("out err=%d\n", err);
	return err;
}



/*
 * rbf_delete_entry deletes a directory entry.
 * Page is up-to-date. Releases the page.
 * 
 * ...should we shrink directory here if possible?
 */
int rbf_delete_entry(struct inode *dir, struct rbf_dirent *de, struct page *page)
{
	struct address_space *mapping = page->mapping;
	unsigned from = (char*)de - (char*)page_address(page);
	unsigned to = from + RBF_DIRENT_SIZE;
	int err;

	lock_page(page);
	err = mapping->a_ops->prepare_write(NULL, page, from, to);
	
	de->dir_name[0] = 0;
	
	err = rbf_commit_chunk(page, from, to);
	unlock_page(page);
	rbf_put_page(page);
	dir->i_ctime = dir->i_mtime = CURRENT_TIME;
	mark_inode_dirty(dir);
	return err;
}


/*
 * Set the first fragment of directory.
 */
int rbf_make_empty(struct inode *dir, struct inode *parent)
{
	struct page *page = rbf_get_page(dir, 0);
	struct rbf_dirent * de;
	unsigned len = RBF_DIRENT_SIZE * 2;
	int err;

	if (page == 0)
		return -ENOMEM;
	
	lock_page(page);
	err = page->mapping->a_ops->prepare_write(NULL, page, 0, len);
	if (err)
		goto fail;

	de = (struct rbf_dirent *)page_address(page);
	
	de[0].dir_name[0] = '.';
	de[0].dir_name[1] = '.' | 0x80;
	rbf_write(de[0].dir_addr, 3, parent->i_ino);
	
	de[1].dir_name[0] = '.' | 0x80;
	rbf_write(de[1].dir_addr, 3, dir->i_ino);
	
	err = rbf_commit_chunk(page, 0, len);
	if (!err) {
		dir->i_size = len;
		mark_inode_dirty(dir);
	}
fail:
	unlock_page(page);
	rbf_put_page(page);

	return err;
}


/*
 * routine to check that the specified directory is empty (for rmdir)
 */
int rbf_empty_dir(struct inode *dir)
{
	struct page *page;
	struct rbf_dirent *de;
	unsigned pos = RBF_DIRENT_SIZE * 2;
	unsigned int offset;
	char *kaddr;

	DPRINTK("in\n");
	
	while (pos < dir->i_size) {
		offset = pos & (PAGE_CACHE_SIZE-1);
		page = rbf_get_page(dir, pos >> PAGE_CACHE_SHIFT);
		
		if (IS_ERR(page)) {
			pos += PAGE_CACHE_SIZE - offset;
			continue;
		}
		kaddr = page_address(page);
		do {
			de = (struct rbf_dirent *)(kaddr + offset);
			if (de->dir_name[0])
				goto not_empty;
			offset += RBF_DIRENT_SIZE;
			pos += RBF_DIRENT_SIZE;
		}
		while (offset < PAGE_CACHE_SIZE && pos < dir->i_size);
		rbf_put_page(page);
	}
	DPRINTK("out 1\n");
	return 1;

not_empty:
	rbf_put_page(page);
	DPRINTK("out 0\n");
	return 0;
}
