/*
 *	linux/fs/rbf/file.c
 *	RBF File System - file operations.
 *	
 *	Copyright (C) 2001 Andrew Cannon <ajc@gmx.net>
 *
 * 	...partly stolen from linux/mm/filemap.c
 * 	
 *  This file is subject to the terms and conditions of the
 *  GNU General Public License. See the file COPYING in the
 *  main directory of this archive for more details.
 */

#include "rbf_defs.h"
#include "rbf_block.h"

#undef DEBUG
#define DEBUG

#ifndef DEBUG
#undef DPRINTK
#define DPRINTK(x...)
#endif

static ssize_t rbf_file_read(struct file *filp, char *buf, size_t count, loff_t *ppos)
{
	struct inode *inode = filp->f_dentry->d_inode;
	struct super_block *s = inode->i_sb;
	struct buffer_head *bh;
	ssize_t retval = 0;
	int offset, len;

//	DPRINTK("in\n");
	
	if ((ssize_t) count < 0)
		return -EINVAL;

//	if (filp->f_flags & O_DIRECT)
//		goto o_direct;

	if (*ppos + count > inode->i_size)
		count = inode->i_size - *ppos;
	
	while (count) {
		bh = rbf_logical_bread(inode, *ppos, &offset);
		if (bh == 0)
			break;
		len = s_su_sectsize - (offset & (s_su_sectsize - 1));
		if (len > count)
			len = count;
		if (copy_to_user(buf, bh->b_data + offset, len))
			return -EFAULT;
		brelse(bh);
		buf += len;
		*ppos += len;
		count -= len;
		retval += len;
	}

//	DPRINTK("out\n");
	return retval;
}

static ssize_t rbf_file_write(struct file *filp, const char *buf, size_t count, loff_t *ppos)
{
	struct inode *inode = filp->f_dentry->d_inode;
	struct super_block *s = inode->i_sb;
	struct buffer_head *bh;
	loff_t		pos;
	ssize_t		ret;
	size_t		zcount;
	int		offset, len, zlen;
	int		nsect;
	long		lsn, psn;

	DPRINTK("in count %d pos %d\n", (int)count, (int)*ppos);
	
	if ((ssize_t) count < 0)
		return -EINVAL;

	if (!access_ok(VERIFY_READ, buf, count))
		return -EFAULT;
	
	if (filp->f_flags & O_APPEND)
		pos = inode->i_size;
	else
		pos = *ppos;

	ret = -EINVAL;
	if (pos < 0)
		goto out;

	/*
	 *	LFS rule 
	 */
	if (pos + count > MAX_NON_LFS) {
		if (pos >= MAX_NON_LFS) {
			send_sig(SIGXFSZ, current, 0);
			goto out;
		}
		if (count > MAX_NON_LFS - (u32)pos) {
			/* send_sig(SIGXFSZ, current, 0); */
			count = MAX_NON_LFS - (u32)pos;
		}
	}

	/*
	 *	Are we about to exceed the fs block limit ?
	 *
	 *	If we have written data it becomes a short write
	 *	If we have exceeded without writing data we send
	 *	a signal and give them an EFBIG.
	 *
	 *	Linus frestrict idea will clean these up nicely..
	 */
	if (pos >= inode->i_sb->s_maxbytes) {
		if (count || pos > inode->i_sb->s_maxbytes) {
			send_sig(SIGXFSZ, current, 0);
			ret = -EFBIG;
			goto out;
		}
		/* zero-length writes at ->s_maxbytes are OK */
	}
	if (pos + count > inode->i_sb->s_maxbytes)
		count = inode->i_sb->s_maxbytes - pos;

	ret = 0;
	if (count == 0)
		goto out;

	inode->i_ctime = inode->i_mtime = CURRENT_TIME;
	mark_inode_dirty_sync(inode);

//	if (filp->f_flags & O_DIRECT)
//		goto o_direct;

	ret = rbf_find_sector(inode, (pos + count - 1) >> s_su_sectsize_bits, 0, 1);	/* expand file */
	if (ret < 0)
		goto out;
	
	if (pos > inode->i_size) {
		zcount = pos - inode->i_size;
		count += zcount;
		pos = inode->i_size;
	}
	else zcount = 0;
	
	ret = 0;
	do {
		if (pos & (s_su_sectsize - 1) || count < s->s_blocksize) {
			bh = rbf_logical_bread(inode, pos, &offset);
			len = s_su_sectsize - (offset & (s_su_sectsize - 1));
			DPRINTK("pos %ld offset %d bh %08lx len %d\n", (long int)pos, offset, (long int)bh, len);
		}
		else	{
			/*
			 *	starting at sector boundary
			 *	
			 *	tricky stuff needed here to discover when a whole
			 *	BUFFER will be written to. Only then can we forego
			 *	the pre-read stage.
			 *	
			 *	1. start must be on block boundary
			 *	2. count must be >= blocksize
			 */
			lsn = pos >> s_su_sectsize_bits;
			psn = rbf_find_sector(inode, lsn, &nsect, 0);
			
			if (psn == 0) {
				printk("%s: lsn %08lx not found at %08lx\n", __FUNCTION__, lsn, (long)pos);
				break;
			}
			else if ((psn & s_su_part_mask) == 0 && nsect > s_su_part_mask) {
				/*
				 *	writing full buffer
				 */
				bh = sb_getblk(s, psn >> s_su_part_shift);
				len = s->s_blocksize;
				offset = 0;
			}
			else {
				bh = rbf_bread(s, psn, &offset);
				len = min((int)s->s_blocksize - offset, nsect << s_su_sectsize_bits);
			}
			DPRINTK("pos %ld lsn %08lx psn %08lx offset %d bh %08lx len %d\n",
					(long int)pos, lsn, psn, offset, (long int)bh, len);
		}
		if (bh == 0) {
			printk("%s: missing buffer at %08x\n", __FUNCTION__, (int)pos);
			break;
		}
		if (len > count)
			len = count;
		if (zcount) {
			zlen = min((size_t)len, zcount);
			DPRINTK("zeroing %d at %d\n", zlen, (int)pos);
			memset(bh->b_data + offset, 0, zlen);
			offset += zlen;
			pos += zlen;
			len -= zlen;
			count -= zlen;
			zcount -= zlen;
		}
		if (len) {
			if (copy_from_user(bh->b_data + offset, buf, len))
				return -EFAULT;
			pos += len;
			buf += len;
			ret += len;
			count -= len;
		}
		rbf_update_buffer(IS_SYNC(inode), bh);
		brelse(bh);
	}
	while (count);
	pos += zcount;		// in case of error while zeroing
	*ppos = pos;
	if (pos > inode->i_size) {
		inode->i_size = pos;
		mark_inode_dirty(inode);
	}
out:	
	DPRINTK("out %d\n", (int)ret);
	
	return ret;
}


struct file_operations rbf_block_file_operations = {
	llseek: 	generic_file_llseek,
	read:		rbf_file_read,
	write:		rbf_file_write,
	release:	rbf_release_file,
};

struct address_space_operations rbf_block_address_space_ops = {
};
